-- 载入一些库
require "sys"
require "bit"
require "lfs"
require "std"
require "ex"
require "ydwelua"
require "titan_engine"
require "YDWELuaMapAnalyzer"
require "sfmpq"
local template_compile = require "template"

local stormlib = ar.stormlib
local sfmpq = ar.sfmpq

local is_inject_slk_file = false
local jasshelper_path = yi.get_editor_root_dir() / "plugin" / "jasshelper"
local wave_path = yi.get_editor_root_dir() / "plugin" / "wave"

-- 载入一个文件的内容
-- file_path - 文件路径, 必须是fs.path类型
-- 返回文件内容, nil表示出错
local function load_file(file_path)
	local f, e = io.open(file_path:string(), "rb")

	if f then
		local content = f:read("*a")
		f:close()
		return content
	else
		return nil, e
	end
end

-- 保存一个文件的内容
-- file_path - 文件路径, 必须是fs.path类型
-- content - 文件内容,必须是字符串
-- 返回true表示成功,false表示失败
local function save_file(file_path, content)
	local f, e = io.open(file_path:string(), "wb")

	if f then
		f:write(content)
		f:close()
		return true
	else
		return false, e
	end
end

-- 根据版本获取YDWE自带的Jass库函数（bj和cj）路径
-- version - 魔兽版本，数
-- 返回：cj路径，bj路径，都是fs.path
local function get_default_jass_libs(version)
	if version >= 1.24 then
		return (yi.get_editor_root_dir() / "jass" / "system" / "ht" / "common.j"),
			(yi.get_editor_root_dir() / "jass" / "system" / "ht" / "blizzard.j")
	else
		return (yi.get_editor_root_dir() / "jass" / "system" / "rb" / "common.j"),
			(yi.get_editor_root_dir() / "jass" / "system" / "rb" / "blizzard.j")
	end
end

-- 准备魔兽争霸3的Jass库函数（common.j和blizzard.j）供语法检查用
-- 如果地图中有，则优先使用地图的，否则使用自带的
-- map_path - 地图路径，fs.path对象
-- 返回2个值：cj路径，bj路径，都是fs.path。
local function prepare_jass_libs(map_path, version)
	local common_j_path = jasshelper_path / "common.j"
	local blizzard_j_path = jasshelper_path / "blizzard.j"
	local map_has_cj = false
	local map_has_bj = false
	
	-- 从地图中解压缩两个重要文件到jasshelper目录（供语法检查用）
	local mpq_handle = stormlib.open_archive(map_path, 0, 0)
	if mpq_handle then
		-- 如果地图中导入了，优先使用地图的
		if stormlib.has_file(mpq_handle, "common.j") then
			stormlib.extract_file(mpq_handle, common_j_path, "common.j")
			map_has_cj = true
		elseif stormlib.has_file(mpq_handle, "scripts\\common.j") then
			stormlib.extract_file(mpq_handle, common_j_path, "scripts\\common.j")
			map_has_cj = true
		end

		if stormlib.has_file(mpq_handle, "blizzard.j") then
			stormlib.extract_file(mpq_handle, blizzard_j_path, "blizzard.j")
			map_has_bj = true
		elseif stormlib.has_file(mpq_handle, "scripts\\blizzard.j") then
			stormlib.extract_file(mpq_handle, blizzard_j_path, "scripts\\blizzard.j")
			map_has_bj = true
		end
		stormlib.close_archive(mpq_handle)
	else
		log.warn("Cannot open map archive, using default bj and cj instead.")
	end

	-- 是否和当前版本一致？
	local use_default = mathx.feq(local_warcraft_version, version, 1e-10)
	local default_common_j_path, default_blizzard_j_path = get_default_jass_libs(version)
	if not map_has_cj then
		if use_default then
			if yi.native_generic_file_exists("common.j") then
				yi.native_generic_extract_file(common_j_path, "common.j")
			elseif yi.native_generic_file_exists("scripts\\common.j") then
				yi.native_generic_extract_file(common_j_path, "scripts\\common.j")
			else			
				common_j_path = default_common_j_path
			end
		else
			common_j_path = default_common_j_path
		end
	end
	if not map_has_bj then
		if use_default then
			if yi.native_generic_file_exists("blizzard.j") then
				yi.native_generic_extract_file(blizzard_j_path, "blizzard.j")
			elseif yi.native_generic_file_exists("scripts\\blizzard.j") then
				yi.native_generic_extract_file(blizzard_j_path, "scripts\\blizzard.j")
			else
				blizzard_j_path = default_blizzard_j_path
			end
		else
			blizzard_j_path = default_blizzard_j_path
		end
	end
	
	return common_j_path, blizzard_j_path
end

-- 启动并等待程序执行完毕，然后检测返回值是否为0
-- command_line - 命令行
-- start_path - 启动时的当前目录
-- 返回：true表明执行成功*并且*目标程序返回值为0，false表明执行失败*或者*目标程序返回值不是0
local function spawn_dir_wait_return(command_line, start_path)
	local success, result = sys.spawn_dir_wait(command_line, start_path)
	if success then
		log.trace(string.format("Executed %s, returned %d", command_line, result))
	else
		log.trace(string.format("Executed %s failed", command_line))
	end
	return success and (result == 0)
end

-- 使用JassHelper编译地图
-- map_path - 地图路径，fs.path对象
-- common_j_path - common.j路径，fs.path对象
-- blizzard_j_path - blizzard.j路径，fs.path对象
-- option - 编译选项, table，目前支持参数：
-- 	enable_jasshelper - 启用JassHelper，true/false
--	enable_jasshelper_debug - 启用JassHelper的Debug，true/false
--	enable_jasshelper_optimization - 启用优化，true/false
-- 返回：true编译成功，false编译失败
local function jasshelper_compile_map(map_path, common_j_path, blizzard_j_path, option)
	-- 确定JassHelper编译器的路径
	local jasshelper_exe_path = jasshelper_path / "jasshelper.exe"
	-- 参数
	local parameter = ""

	-- 需要做vJass编译？
	if option.enable_jasshelper then
		-- debug选项（--debug）
		if option.enable_jasshelper_debug then
			parameter = parameter .. " --debug"
		end
		-- （关闭）优化选项（--nooptimize）
		if not option.enable_jasshelper_optimization then
			parameter = parameter .. " --nooptimize"
		end
	else
		-- 不编译vJass选项（--nopreprocessor）
		parameter = parameter .. " --nopreprocessor"
	end

	-- 生成命令行
	local command_line = string.format('"%s"%s "%s" "%s" "%s"',
		jasshelper_exe_path:string(),
		parameter,
		common_j_path:string(),
		blizzard_j_path:string(),
		map_path:string()
	)

	-- 执行并获取结果
	return spawn_dir_wait_return(command_line, yi.get_editor_root_dir())
end

-- 使用cJass编译地图
-- map_path - 地图路径，fs.path对象
-- option - 附加编译选项, table，支持选项为：
--	enable_jasshelper_debug - 启用Debug模式，true/false
--	runtime_version - 魔兽版本，数
-- 返回：true编译成功，false编译失败
local function cjass_compile_map(map_path, option)
	-- 确定cJass编译器的路径
	local cjass_path = yi.get_editor_root_dir() / "plugin" / "AdicHelper" / "AdicHelper.exe"

	-- 确定附加参数
	local parameter = option.runtime_version >= 1.24 and " /v24" or " /v23"
	parameter = parameter .. (option.enable_jasshelper_debug and " /dbg" or "")

	-- 命令行
	local command_line = string.format('"%s"%s /mappars="%s"',
		cjass_path:string(),
		parameter,
		map_path:string()
	)

	-- 执行并获取结果
	return spawn_dir_wait_return(command_line, yi.get_editor_root_dir() / "plugin" / "AdicHelper")
end

-- 注入代码到Jass代码文件（最常见的是war3map.j）中
-- map_script_path - war3map.j的路径，fs.path对象
-- inject_code_path_table - 所有需要注入的代码文件路径，table，table中可以是
-- 		string - 此时为YDWE / "jass" 目录下的对应名称的文件
--		fs.path - 此时取其路径
-- 注：该table必须是数组形式的，哈希表形式的不处理
-- 返回值：0 - 成功；-1 - 出错失败；1 - 什么都没做
local function inject_code_to_jass_script(map_script_path, inject_code_path_table)
	-- 结果
	local result = 1
	if inject_code_path_table and #inject_code_path_table > 0 then
		-- 默认成功
		result = 0
		log.trace("Writing code to " .. map_script_path:filename():string())

		-- 打开文件供写入（追加模式）
		local map_script_file, e = io.open(map_script_path:string(), "ab+")
		if map_script_file then

			--if global_config:get_integer("ScriptCompiler.EnableCJass", 0) == 0 then
			--	map_script_file:write("#include <YDTrigger/LeakMonitorMacro.h>\r\n")
			--end

			-- 循环处理每个需要注入的文件
			for index, inject_code_path in ipairs(inject_code_path_table) do
				local inject_code_path_string = nil
				if type(inject_code_path) == "string" then
					if inject_code_path:find("\\") or inject_code_path:find("/") then
						inject_code_path_string = inject_code_path
					else
						inject_code_path_string = (yi.get_editor_root_dir() / "jass" / inject_code_path):string()
					end
				else
					inject_code_path_string = inject_code_path:string()
				end
				-- 打开代码文件
				local f, e = io.open(inject_code_path_string, "rb")
				if f then
					log.trace("Injecting " .. inject_code_path_string)
					-- 读取所有内容
					local code_content = f:read("*a")
					-- 关闭文件
					f:close()

					-- 插入代码到原文件最后
					map_script_file:write(code_content)
					-- 写上一个换行符（我日他妈的cJass你居然不支持Linux格式的换行符）
					map_script_file:write("\r\n")
					-- 成功
					log.trace("Injection completed")
				else
					result = -1
					log.error("Error occured when reading code to inject.")
					log.error(e)
				end
			end

			--if global_config:get_integer("ScriptCompiler.EnableCJass", 0) == 0 then
			--	map_script_file:write("#include <YDTrigger/LeakMonitorMacro.h>\r\n")
			--end

			-- 关闭文件
			map_script_file:close()
		else
			result = -1
			log.error("Error occured when writing code to map script")
			log.error(e)
		end
	end

	return result
end

-- 确定应当把地图保存为适合老版本（< 1.24）还是新版本（>= 1.24）的
-- 对于保存为双份，按照当前魔兽版本处理
-- save_type - 保存类型，0到3的取值，意义同设置程序，从上到下依次为0到3
-- runtime_version - 魔兽版本，数
-- 返回值: 数，如果保存的目标版本和当前版本一致，返回当前版本，否则视情况返回1.20或者1.24（和当前版本不同）
local function determine_map_version(save_type, runtime_version)
	-- 固定旧版本
	if save_type == 1 then
		-- 如果当前版本也是旧版本，返回当前版本
		if runtime_version < 1.24 then
			return runtime_version
		else
			-- 否则返回一个旧版本的典型代表（1.20）
			return 1.20
		end
	-- 固定新版本
	elseif save_type == 2 then
		-- 如果当前版本也是新版本，返回当前版本
		if runtime_version >= 1.24 then
			return runtime_version
		else
			-- 否则返回一个新版本的典型代表（1.24）
			return 1.24
		end
	else
		-- 按照当前版本或者双份
		return runtime_version
	end
end


-- 侦测需要注入哪些代码
-- map_script_path - 脚本的路径，fs.path变量
-- option - 选项，table类型，支持成员：
-- 	runtime_version - 数，表示魔兽版本
-- 返回：一个table，数组形式，包含所有需要注入的文件名（注意不是fs.path）
local function detect_code_injection(map_script_path, option)
	-- 结果变量
	local result = nil

	-- 读入所有文本
	local f, e = io.open(map_script_path:string(), "rb")
	-- 文件存在
	if f then
		-- 读入所有东西
		local s = f:read("*a")
		-- 结果
		result = {}

		-- 检查是否有需要注入的函数
		local inject_code_table = option.runtime_version >= 1.24 and inject_code_table_new or inject_code_table_old

		local GeneralBounsSystemFile = yi.get_editor_root_dir() / "jass" / "YDWEGeneralBounsSystem.j"

		for file, function_table in pairs(inject_code_table) do
		    if GeneralBounsSystemFile:string() == file:string() then
				for i, function_name in ipairs(function_table) do
					if s:find(function_name) then
					    is_inject_slk_file = true
						table.insert(result, file)
						break
					end
				end
			else
				for i, function_name in ipairs(function_table) do
					if s:find(function_name) then
						table.insert(result, file)
						break
					end
				end
			end
		end

		-- 关闭文件
		f:close()
	else
		log.error("Error occured when opening map script.")
		log.error(e)
	end

	return result
end

local map_file_import = function(filename)
	return function (buf)
		local temp_file_path = yi.get_editor_root_dir() / "logs" / "import" / filename
		fs.create_directories(temp_file_path:parent_path())
		if save_file(temp_file_path, buf) then
			if ar.stormlib.add_or_replace_file(
				__map_handle__,
				temp_file_path,
				filename,
				ydwelua.MPQ_FILE_REPLACEEXISTING
			) then
				log.trace("import file succeeded.")
				return
			end
		end
		log.error("import file failed.")
	end
end

local function lua_processed_map_script(map_path, map_handle, map_script_path, out_script_path)
	log.trace("Lua template processing: map_path - " .. map_path:string())
	local result = false
	local content, err = load_file(map_script_path)
	if content then
		__map_handle__ = map_handle
		local success, content = template_compile(content, setmetatable({import = map_file_import}, {__index = _G}))
		if success then
			result = save_file(out_script_path, content)
			log.debug("Process completed")
		else
			if content then
				sys.message_dialog(nil, content, _("Error"), bit.bor(ydwelua.MB_ICONERROR, ydwelua.MB_OK))
			else
				sys.message_dialog(nil, _("Unknown"), _("Error"), bit.bor(ydwelua.MB_ICONERROR, ydwelua.MB_OK))
			end
			
			log.warn("Error processing: " .. tostring(content))
		end
	else
		log.warn("Cannot load " .. map_script_path:string() .. ". Error: " .. err)
	end

	return result
end

-- 使用管道方式打开一个进程
local function pipe_open(...)
	local in_rd, in_wr = io.pipe()
	local out_rd, out_wr = io.pipe()
	local err_rd, err_wr = io.pipe()
	local proc, err = os.spawn{stdin = in_rd, stdout = out_wr, stderr = err_wr, ...}

	-- 关闭管道不用的一端
	in_rd:close()
	out_wr:close()
	err_wr:close()
	if proc then
		return proc, out_rd, err_rd, in_wr
	else
		in_wr:close()
		out_rd:close()
		err_rd:close()
		return proc, err, nil, nil
	end
end


-- 预处理代码
-- code_path - 输入文件路径
-- option - 预处理选项，table，支持的值有
-- 	runtime_version - 数，表示魔兽版本
-- 	enable_jasshelper_debug - 布尔值，是否是调试模式
--	enable_yd_trigger - 布尔值，是否启用YD触发器
-- 返回：number, info, path - 子进程返回值；预处理输出信息；输出文件路径
local function preprocessor(code_path, option)
	-- Wave的路径
	local wave_exe_path = wave_path / "Wave.exe"
	-- 配置文件的路径
	local wave_config_file_path = wave_path / "WaveConfig.cfg"
	-- Jass文件路径
	local jass_dir_path = yi.get_editor_root_dir() / "jass"
	-- 输出路径
	local out_file_path = code_path:parent_path() / (code_path:stem():string() .. ".i")
	-- Wave系统包含路径
	local wave_sysinclude_dir = wave_path / "include"
	-- 强制包含文件路径
	local force_file_path = wave_sysinclude_dir / "WaveForce.i"

	-- 生成配置文件
	local f, e = io.open(wave_config_file_path:string(), "w")
	if f then
		-- 写入内容
		f:write("--autooutput\n")
		f:write(string.format('--sysinclude="%s"\n', wave_sysinclude_dir:string()))
		f:write(string.format('--sysinclude="%s"\n', (yi.get_editor_root_dir() / "plugin"):string()))
		f:write(string.format('--include="%s"\n', jass_dir_path:string()))
		f:write(string.format('--define=WARCRAFT_VERSION=%d\n', 100 * option.runtime_version))
		f:write(string.format('--define=YDWE_VERSION_STRING="%s"\n', yd_version.full_string))
		if option.enable_jasshelper_debug then
			f:write('--define=DEBUG=1\n')
		end
		if global_config:get_integer("ScriptInjection.Option", 0) == 0 then
			f:write("--define=SCRIPT_INJECTION=1\n")
		end
		if not option.enable_yd_trigger then
			f:write('--define=DISABLE_YDTRIGGER=1\n')
		end
		if fs.exists(force_file_path) then
			f:write(string.format('--forceinclude=%s\n', force_file_path:filename():string()))
		end
		f:write("--extended\n--c99\n--preserve=2\n--line=0\n")
		f:close()

		-- 启动进程
		local proc, out_rd, err_rd, in_wr = pipe_open(
			string.format('"%s"', wave_exe_path:string()),
			string.format('@"%s"', wave_config_file_path:string()),
			string.format('"%s"', code_path:string())
		)
		if proc then
			local out = out_rd:read("*a")
			local err = err_rd:read("*a")
			local exit_code = proc:wait()
			return exit_code, out, err, out_file_path
		else
			return -1, nil, nil, out_file_path
		end
	else
		return -1, nil, nil, out_file_path
	end
end


-- 添加文件给地图
-- map_path - 地图路径，fs.path
-- file_path - 需要加入的文件的路径，fs.path
-- path_in_archive - 地图压缩包中的路径，string
-- 返回值：true表示成功，false表示失败
local function insert_map_archive_file(map_path, file_path, path_in_archive)
	map_path = (type(map_path) == "userdata") and map_path:string() or map_path
	-- 结果
	local result = false

	log.trace("Add " .. path_in_archive)

	-- 打开MPQ（地图）
	local mpq_handle = sfmpq.MpqOpenArchiveForUpdate(map_path, 4, 0)
	if mpq_handle ~= 0 then
		if 0 ~= sfmpq.MpqAddFileToArchiveEx(
			mpq_handle,
			file_path:string(),
			path_in_archive,
			0x00201,
			2,
			9
		) then
			result = true
		end

		-- 关闭地图
		sfmpq.MpqCloseUpdatedArchive(mpq_handle, 0)
	else
		log.error("Cannot open map archive " .. map_path)
	end

	return result
end

-- 添加文件给地图
-- map_path - 地图路径，fs.path
-- path_in_archive - 地图压缩包中的路径，string
-- 返回值：true表示成功，false表示失败
local function add_map_archive_file(map_path, path_in_archive)
	local result = false
	local extract_file_path = yi.get_editor_root_dir() / "logs" / path_in_archive
	fs.create_directories(extract_file_path:parent_path())
	if yi.native_generic_extract_file(extract_file_path, path_in_archive) then
		result = insert_map_archive_file(map_path, extract_file_path, path_in_archive)

		-- 删除临时文件
		pcall(fs.remove_all, extract_file_path)
	else
		log.error("Cannot extract " .. path_in_archive)
	end
	return result
end

-- 从地图中解压出文件来然后调用回调函数更新
-- map_path - 地图路径，fs.path
-- path_in_archive - 地图压缩包中的路径，string
-- process_function - 函数，必须接收一个fs.path对象，返回一个fs.path对象
-- 形如 function (in_path) return out_path end
-- 返回值：true表示成功，false表示失败
local function update_map_archive_file(map_path, path_in_archive, process_function)
	-- 结果
	local result = false
	log.trace("Updating " .. map_path:filename():string())

	-- 打开MPQ（地图）
	local mpq_handle = stormlib.open_archive(map_path, 0, 0)
	if mpq_handle then
		-- 确定解压路径
		local extract_file_path = yi.get_editor_root_dir() / "logs" / "file.out"
		-- 将文件解压
		if stormlib.has_file(mpq_handle, path_in_archive) and
			stormlib.extract_file(mpq_handle, extract_file_path, path_in_archive)
		then
			log.trace(path_in_archive .. " has been extracted from " .. map_path:filename():string())

			-- 调用处理函数处理
			local success, out_file_path = pcall(process_function, mpq_handle, extract_file_path)
			-- 如果函数正常结束（没有出错）
			if success then
				-- 如果函数成功完成任务
				if out_file_path then
					-- 替换文件
					if stormlib.add_or_replace_file(
						mpq_handle,
						out_file_path,
						path_in_archive,
						ydwelua.MPQ_FILE_REPLACEEXISTING
					) then
						log.trace("Archive update succeeded.")
						result = true
					else
						log.error("Error occurred when write back")
					end
				else
					-- 出现了错误--
					log.error("Processor function cannot complete its task.")
				end
			else
				-- 记录出错原因
				log.error(out_file_path)
			end

			-- 删除临时文件
			pcall(fs.remove_all, extract_file_path)
		else
			log.error("Cannot extract " .. path_in_archive)
		end

		-- 关闭地图
		stormlib.close_archive(mpq_handle)
	else
		log.error("Cannot open map archive" .. map_path:string())
	end

	return result
end

-- 给地图名末尾，扩展名前添加内容
-- 效果：("abc.w3x", "def") -> "abcdef.w3x"
-- map_path_string - 地图文件名或者路径
-- some_string - 需要添加的字符串
-- 返回：字符串，见效果
local function generate_aux_map_path_string_of_string(map_path_string, some_string)
	if #map_path_string >= 4 then
		return map_path_string:sub(1, -4) .. some_string .. map_path_string:sub(-4)
	else
		return some_string .. map_path_string
	end
end

-- 功能同上，但第一个参数是fs.path
local function generate_aux_map_path_string_of_path(map_path, some_string)
	return generate_aux_map_path_string_of_string(map_path:string(), some_string)
end

-- 功能同上，但第一个参数是fs.path，并且返回fs.path
local function generate_aux_map_path_of_path(map_path, some_string)
	return fs.path(generate_aux_map_path_string_of_path(map_path, some_string))
end

-- 编译地图
-- map_path - 地图路径，string或者fs.path
-- option - 编译选项，table类型，支持的选项：
--	enable_preprocessor - 启用wave预处理器，true/false
-- 	enable_jasshelper - 启用JassHelper，true/false
--	enable_jasshelper_debug - 启用JassHelper的Debug，true/false
--	enable_jasshelper_optimization - 启用优化，true/false
--	enable_cjass - 启用cJass编译器，true/false
--	script_injection - 代码注入选项，整数，0表示智能注入，1表示不注入
--	runtime_version - 魔兽版本，数
-- 返回：true表示成功，false表示失败
local function compile_map(map_path, option)
	-- 统一处理fs.path
	if type(map_path) == "string" then map_path = fs.path(map_path) end
	-- 结果
	local result = nil

	-- 准备common.j以及blizzard.j
	local common_j_path, blizzard_j_path = prepare_jass_libs(map_path, option.runtime_version)

	-- 如果JassHelper开启，执行正常编译
	if option.enable_jasshelper then
		-- 根据注入类型决定注入内容
		is_inject_slk_file = false

		result = update_map_archive_file(map_path, "war3map.j",
			-- 解压缩地图脚本，处理然后写回
			function (map_handle, in_script_path)
				-- 开始处理
				log.trace("Processing " .. in_script_path:filename():string())
				local lua_processed_script_path = yi.get_editor_root_dir() / "logs" / "lua_processed.j"

				-- 未启用用cJass
				if not option.enable_cjass then
					-- 根据注入选项进行处理（由于Lua的closure，此处可以访问“父”函数的局部变量）
					if option.script_injection == 0 then
						inject_code_to_jass_script(
							in_script_path,
							detect_code_injection(in_script_path, option)
						)
					end
	
					log.trace("Preprocessing " .. in_script_path:filename():string())	
					-- Wave预处理
					local exit_code, out, err, preprocessed_path = preprocessor(in_script_path, option)
	
					-- 退出码0代表成功
					if exit_code ~= 0 then
						if out and err then
							-- 显示出错信息
							local message = string.format(_("Preprocessor failed with message:\nstdout:%s\nstderr: %s"), out, err)
							sys.message_dialog(nil, message, _("Error"), bit.bor(ydwelua.MB_ICONERROR, ydwelua.MB_OK))
						else
							sys.message_dialog(nil, _("Cannot start preprocessor process."), _("Error"), bit.bor(ydwelua.MB_ICONERROR, ydwelua.MB_OK))
						end
						return nil
					end
					
					in_script_path = preprocessed_path
				end
				
				-- Lua预处理
				assert(lua_processed_map_script(map_path, map_handle, in_script_path, lua_processed_script_path), "Lua template failed")
				return lua_processed_script_path
			end
		)

		if is_inject_slk_file then
			add_map_archive_file(map_path, "units\\abilitydata.slk")
			--会掉线
			--add_map_archive_file(map_path, "units\\abilitymetadata.slk")
		end
		
		-- 开始调用编译工具编译
		if result then
			-- 调用cJass
			if option.enable_cjass then
				result = cjass_compile_map(map_path, option)
				if result then
					log.debug("CJass compilation succeeded.")
				end
			end

			-- 调用jasshelepr
			if result then
				result = jasshelper_compile_map(map_path, common_j_path, blizzard_j_path, option)
				if result then
					log.debug("JassHelper compilation succeeded.")
				end
			else
				log.warn("CJass compilation failed.")
			end
		else
			log.warn("Early processing stage completed with error.")
		end
	else
		-- 只做语法检查
		result = jasshelper_compile_map(map_path, common_j_path, blizzard_j_path, option)
	end

	return result
end


-- 本函数当保存地图时调用
-- event_data - 事件参数。table类型，包含了以下成员
--	map_path - 保存的地图路径，字符串类型
-- 返回值：0和正数表示保存成功，负数表示保存失败。如果成功编译，一般返回0
function on_save(event_data)
	log.debug("********************* on save start *********************")

	-- 刷新配置数据
	reload_configuration()

	-- 当前保存的地图路径
	local map_path = fs.path(event_data.map_path)
	log.trace("Saving " .. map_path:string())

	-- 获取保存类型
	local save_type = global_config:get_integer("MapSave.Option", 0)

	-- 获取保存选项
	local save_option = {
		-- 是否启用JassHelper
		enable_jasshelper = (global_config:get_integer("ScriptCompiler.EnableJassHelper", 1) == 1),
		-- 是否是调试模式
		enable_jasshelper_debug = (global_config:get_integer("ScriptCompiler.EnableJassHelperDebug", 0) == 1),
		-- 是否优化地图
		enable_jasshelper_optimization = (global_config:get_integer("ScriptCompiler.EnableJassHelperOptimization", 1) == 1),
		-- 是否启用cJass
		enable_cjass = (global_config:get_integer("ScriptCompiler.EnableCJass", 0) == 1),
		-- 目标魔兽版本
		runtime_version = determine_map_version(save_type, local_warcraft_version),
		-- 代码注入选项
		script_injection = global_config:get_integer("ScriptInjection.Option", 0),
		-- 是否启用YDTrigger
		enable_yd_trigger = plugin.loaders['YDTrigger'] ~= nil
	}

	-- cjass和脚本注入不能共存
	if save_option.enable_cjass then
		save_option.script_injection = false
	end

	-- 如果保存双份，需要预先拷贝
	local map_path_aux = nil
	if save_type == 3 then
		-- 决定第二份的路径
		if save_option.runtime_version < 1.24 then
			map_path_aux = generate_aux_map_path_of_path(map_path, "hashtable")
		else
			map_path_aux = generate_aux_map_path_of_path(map_path, "returnbug")
		end

		log.trace("Making copy of the original map.")

		-- true代表覆盖现有文件。false不会覆盖
		pcall(fs.copy_file, map_path, map_path_aux, true)
	end

	-- 编译地图
	local result = compile_map(map_path, save_option)

	-- 如果有第二份，此时处理第二份
	if map_path_aux then
		log.trace("Processing second map")
		save_option.runtime_version = save_option.runtime_version >= 1.24 and 1.20 or 1.24
		compile_map(map_path_aux, save_option)
	end

	log.debug("********************* on save end *********************")

	if result then return 0 else return -1 end
end
